# Copyright (c) Sean Vig 2020
# Copyright (c) Matt Colligan 2021

from __future__ import annotations

import typing

from wlroots import Ptr, ffi, lib

if typing.TYPE_CHECKING:
    from wlroots.renderer import Renderer
    from wlroots.util.region import PixmanRegion32
    from wlroots.wlr_types.buffer import Buffer


class Texture(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ptr

    @staticmethod
    def from_pixels(
        renderer: Renderer,
        fmt: int,
        stride: int,
        width: int,
        height: int,
        data: ffi.CData,
    ) -> Texture:
        """
        Create a new texture from raw pixel data. `stride` is in bytes. The returned
        texture is mutable.

        Should not be called in a rendering block like renderer_begin()/end() or
        between attaching a renderer to an output and committing it.
        """
        ptr = lib.wlr_texture_from_pixels(
            renderer._ptr, fmt, stride, width, height, data
        )
        ptr = ffi.gc(ptr, lib.wlr_texture_destroy)
        return Texture(ptr)

    @staticmethod
    def from_buffer(
        renderer: Renderer,
        buffer: Buffer,
    ) -> Texture:
        """Create a new texture from a wlr_buffer."""
        ptr = lib.wlr_texture_from_buffer(renderer._ptr, buffer._ptr)
        ptr = ffi.gc(ptr, lib.wlr_texture_destroy)
        return Texture(ptr)

    def update_from_buffer(
        self,
        buffer: Buffer,
        damage: PixmanRegion32 | None = None,
    ) -> bool:
        """
        Update a texture with a struct wlr_buffer's contents.

        The damage can be used by the renderer as an optimization: only the supplied
        region needs to be updated.
        """
        if damage is None:
            damage_ptr = ffi.NULL
        else:
            damage_ptr = damage._ptr
        return lib.wlr_texture_update_from_buffer(self._ptr, buffer._ptr, damage_ptr)

    def destroy(self) -> None:
        """Destroys this wlr_texture."""
        if self._ptr is not None:
            ffi.release(self._ptr)
            self._ptr = None



# Copyright (c) Sean Vig 2019
# Copyright (c) Matt Colligan 2022

from __future__ import annotations

import enum
from weakref import WeakKeyDictionary

from wlroots import Ptr, PtrHasData, ffi, lib, str_or_none

from .input_device import ButtonState, InputDevice

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


@enum.unique
class AxisSource(enum.IntEnum):
    WHEEL = lib.WLR_AXIS_SOURCE_WHEEL
    FINGER = lib.WLR_AXIS_SOURCE_FINGER
    CONTINUOUS = lib.WLR_AXIS_SOURCE_CONTINUOUS
    WHEEL_TILT = lib.WLR_AXIS_SOURCE_WHEEL_TILT


@enum.unique
class AxisOrientation(enum.IntEnum):
    VERTICAL = lib.WLR_AXIS_ORIENTATION_VERTICAL
    HORIZONTAL = lib.WLR_AXIS_ORIENTATION_HORIZONTAL


class Pointer(PtrHasData):
    def __init__(self, ptr) -> None:
        self._ptr = ptr

    @staticmethod
    def from_input_device(input_device: InputDevice) -> Pointer:
        return Pointer(lib.wlr_pointer_from_input_device(input_device._ptr))

    @property
    def base(self) -> InputDevice:
        device_ptr = ffi.addressof(self._ptr.base)
        _weakkeydict[device_ptr] = self._ptr
        return InputDevice(device_ptr)

    @property
    def output_name(self) -> str | None:
        """The name of any associated output"""
        return str_or_none(self._ptr.output_name)


class _PointerEvent(Ptr):
    @property
    def pointer(self) -> Pointer:
        """The pointer associated with the event"""
        return Pointer(self._ptr.pointer)

    @property
    def time_msec(self) -> int:
        return self._ptr.time_msec


class PointerMotionEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        """A relative motion pointer event"""
        self._ptr = ffi.cast("struct wlr_pointer_motion_event *", ptr)

    @property
    def delta_x(self) -> float:
        return self._ptr.delta_x

    @property
    def delta_y(self) -> float:
        return self._ptr.delta_y

    @property
    def unaccel_delta_x(self) -> float:
        return self._ptr.unaccel_dx

    @property
    def unaccel_delta_y(self) -> float:
        return self._ptr.unaccel_dy


class PointerMotionAbsoluteEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        """A absolute motion pointer event"""
        self._ptr = ffi.cast("struct wlr_pointer_motion_absolute_event *", ptr)

    @property
    def x(self) -> float:
        return self._ptr.x

    @property
    def y(self) -> float:
        return self._ptr.y


class PointerButtonEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        """A pointer button event"""
        self._ptr = ffi.cast("struct wlr_pointer_button_event *", ptr)

    @property
    def button(self) -> int:
        return self._ptr.button

    @property
    def button_state(self) -> ButtonState:
        return ButtonState(self._ptr.state)


class PointerAxisEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        """A pointer axis event"""
        self._ptr = ffi.cast("struct wlr_pointer_axis_event *", ptr)

    @property
    def source(self) -> AxisSource:
        return AxisSource(self._ptr.source)

    @property
    def orientation(self) -> AxisOrientation:
        return AxisOrientation(self._ptr.orientation)

    @property
    def delta(self) -> float:
        return self._ptr.delta

    @property
    def delta_discrete(self) -> int:
        return self._ptr.delta_discrete


class PointerSwipeBeginEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_pointer_swipe_begin_event *", ptr)

    @property
    def fingers(self) -> int:
        return self._ptr.fingers


class PointerSwipeUpdateEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_pointer_swipe_update_event *", ptr)

    @property
    def fingers(self) -> int:
        return self._ptr.fingers

    @property
    def dx(self) -> float:
        return self._ptr.dx

    @property
    def dy(self) -> float:
        return self._ptr.dy


class PointerSwipeEndEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        ptr = ffi.cast("struct wlr_pointer_swipe_end_event *", ptr)
        self._ptr = ptr

    @property
    def cancelled(self) -> bool:
        return self._ptr.cancelled


class PointerPinchBeginEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_pointer_pinch_begin_event *", ptr)

    @property
    def fingers(self) -> int:
        return self._ptr.fingers

    @property
    def time_msec(self) -> int:
        return self._ptr.time_msec

    @property
    def dx(self) -> float:
        return self._ptr.dx

    @property
    def dy(self) -> float:
        return self._ptr.dy


class PointerPinchUpdateEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_pointer_pinch_update_event *", ptr)

    @property
    def fingers(self) -> int:
        return self._ptr.fingers

    @property
    def dx(self) -> float:
        return self._ptr.dx

    @property
    def dy(self) -> float:
        return self._ptr.dy

    @property
    def scale(self) -> float:
        return self._ptr.scale

    @property
    def rotation(self) -> float:
        return self._ptr.rotation

    @property
    def time_msec(self) -> int:
        return self._ptr.time_msec

    @property
    def cancelled(self) -> bool:
        return self._ptr.cancelled


class PointerPinchEndEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        ptr = ffi.cast("struct wlr_pointer_pinch_end_event *", ptr)
        self._ptr = ptr

    @property
    def cancelled(self) -> bool:
        return self._ptr.cancelled


class PointerHoldBeginEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_pointer_hold_begin_event *", ptr)

    @property
    def fingers(self) -> int:
        return self._ptr.fingers


class PointerHoldEndEvent(_PointerEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_pointer_hold_end_event *", ptr)

    @property
    def cancelled(self) -> bool:
        return self._ptr.cancelled



# Copyright (c) Sean Vig 2019
from __future__ import annotations

import enum
from typing import TYPE_CHECKING

from pywayland.server import Signal

from wlroots import PtrHasData, ffi, lib, ptr_or_null

from .compositor import Surface
from .input_device import InputDevice, InputDeviceType
from .output import Output
from .output_layout import OutputLayout
from .pointer import (
    PointerAxisEvent,
    PointerButtonEvent,
    PointerHoldBeginEvent,
    PointerHoldEndEvent,
    PointerMotionAbsoluteEvent,
    PointerMotionEvent,
    PointerPinchBeginEvent,
    PointerPinchEndEvent,
    PointerPinchUpdateEvent,
    PointerSwipeBeginEvent,
    PointerSwipeEndEvent,
    PointerSwipeUpdateEvent,
)
from .touch import (
    TouchCancelEvent,
    TouchDownEvent,
    TouchMotionEvent,
    TouchUpEvent,
)

if TYPE_CHECKING:
    from .xcursor_manager import XCursorManager


class WarpMode(enum.Enum):
    Layout = enum.auto()
    LayoutClosest = enum.auto()
    AbsoluteClosest = enum.auto()


def _ensure_attachable(input_device: InputDevice) -> None:
    """Helper for cursor.attach_input_device and detach_input_device.

    :raises ValueError: If the input device is not allowed.
    """
    allowed_device_types = (
        InputDeviceType.POINTER,
        InputDeviceType.TOUCH,
        InputDeviceType.TABLET_TOOL,
    )
    if input_device.type not in allowed_device_types:
        raise ValueError(
            f"Input device must be one of pointer, touch, or tablet tool, got: {input_device.type.name}"
        )


class Cursor(PtrHasData):
    def __init__(self, output_layout: OutputLayout) -> None:
        """Manage a cursor attached to the given output layout

        Uses the given layout to establish the boundaries and movement
        semantics of this cursor. Cursors without an output layout allow
        infinite movement in any direction and do not support absolute input
        events.

        :param output_layout:
            The output layout to attach the cursor to and bound the cursor by.
        """
        ptr = lib.wlr_cursor_create()
        self._ptr = ffi.gc(ptr, lib.wlr_cursor_destroy)
        lib.wlr_cursor_attach_output_layout(self._ptr, output_layout._ptr)

        self.motion_event = Signal(
            ptr=ffi.addressof(self._ptr.events.motion), data_wrapper=PointerMotionEvent
        )
        self.motion_absolute_event = Signal(
            ptr=ffi.addressof(self._ptr.events.motion_absolute),
            data_wrapper=PointerMotionAbsoluteEvent,
        )
        self.button_event = Signal(
            ptr=ffi.addressof(self._ptr.events.button), data_wrapper=PointerButtonEvent
        )
        self.axis_event = Signal(
            ptr=ffi.addressof(self._ptr.events.axis), data_wrapper=PointerAxisEvent
        )
        self.frame_event = Signal(ptr=ffi.addressof(self._ptr.events.frame))
        self.swipe_begin = Signal(
            ptr=ffi.addressof(self._ptr.events.swipe_begin),
            data_wrapper=PointerSwipeBeginEvent,
        )
        self.swipe_update = Signal(
            ptr=ffi.addressof(self._ptr.events.swipe_update),
            data_wrapper=PointerSwipeUpdateEvent,
        )
        self.swipe_end = Signal(
            ptr=ffi.addressof(self._ptr.events.swipe_end),
            data_wrapper=PointerSwipeEndEvent,
        )
        self.pinch_begin = Signal(
            ptr=ffi.addressof(self._ptr.events.pinch_begin),
            data_wrapper=PointerPinchBeginEvent,
        )
        self.pinch_update = Signal(
            ptr=ffi.addressof(self._ptr.events.pinch_update),
            data_wrapper=PointerPinchUpdateEvent,
        )
        self.pinch_end = Signal(
            ptr=ffi.addressof(self._ptr.events.pinch_end),
            data_wrapper=PointerPinchEndEvent,
        )
        self.hold_begin = Signal(
            ptr=ffi.addressof(self._ptr.events.hold_begin),
            data_wrapper=PointerHoldBeginEvent,
        )
        self.hold_end = Signal(
            ptr=ffi.addressof(self._ptr.events.hold_end),
            data_wrapper=PointerHoldEndEvent,
        )
        self.touch_up_event = Signal(
            ptr=ffi.addressof(self._ptr.events.touch_up),
            data_wrapper=TouchUpEvent,
        )
        self.touch_down_event = Signal(
            ptr=ffi.addressof(self._ptr.events.touch_down),
            data_wrapper=TouchDownEvent,
        )
        self.touch_motion_event = Signal(
            ptr=ffi.addressof(self._ptr.events.touch_motion),
            data_wrapper=TouchMotionEvent,
        )
        self.touch_cancel_event = Signal(
            ptr=ffi.addressof(self._ptr.events.touch_cancel),
            data_wrapper=TouchCancelEvent,
        )
        self.touch_frame_event = Signal(
            ptr=ffi.addressof(self._ptr.events.touch_frame),
        )

    @property
    def x(self) -> float:
        """The x position of the cursor"""
        return self._ptr.x

    @property
    def y(self) -> float:
        """The y position of the cursor"""
        return self._ptr.y

    def destroy(self) -> None:
        """Clean up the cursor"""
        if self._ptr is not None:
            ffi.release(self._ptr)
            self._ptr = None

    def attach_input_device(self, input_device: InputDevice) -> None:
        """Attaches this input device to this cursor

        The input device type must be one of:
            - InputDeviceType.POINTER
            - InputDeviceType.TOUCH
            - InputDeviceType.TABLET_TOOL

        :param input_device:
            The input device to attach to the cursor
        """
        _ensure_attachable(input_device)
        lib.wlr_cursor_attach_input_device(self._ptr, input_device._ptr)

    def detach_input_device(self, input_device: InputDevice) -> None:
        """Detaches the provided input device from this cursor

        See attach_input_device() for allowed types of input devices.

        :param input_device:
            The input device to detach from the cursor.
        """
        _ensure_attachable(input_device)
        lib.wlr_cursor_detach_input_device(self._ptr, input_device._ptr)

    def move(
        self,
        delta_x: float,
        delta_y: float,
        *,
        input_device: InputDevice | None = None,
    ) -> None:
        """Move the cursor in the direction of the given x and y layout coordinates

        If one coordinate is NAN, it will be ignored.

        The `input_device` may be passed to respect device mapping constraints.
        If `input_device` is None, device mapping constraints will be ignored.
        """
        input_device_ptr = ptr_or_null(input_device)
        lib.wlr_cursor_move(self._ptr, input_device_ptr, delta_x, delta_y)

    def warp(
        self,
        warp_mode: WarpMode,
        x: float | None,
        y: float | None,
        *,
        input_device: InputDevice | None = None,
    ) -> bool:
        """Warp the cursor to the given x and y in location

        In Layout and LayoutClosest modes, warp the cursor to the given x and y
        in layout coordinates.  In AbsoluteClosest mode, warp the cursor to the
        given x and y in absolute 0..1 coordinates.

        For Layout mode, if x and y are out of the layout boundaries or
        constraints, no warp will happen.  For LayoutClosest and
        AbsoluteClosest modes, if the given point is out of the layout
        boundaries or constraints, the closest point will be used.

        If one coordinate is None, it will be ignored.

        The `input_device` may be passed to respect device mapping constraints.
        If `input_device` is None, device mapping constraints will be ignored.
        """
        if x is None:
            x = float("NaN")
        if y is None:
            y = float("NaN")

        input_device_ptr = ptr_or_null(input_device)

        if warp_mode == WarpMode.Layout:
            return lib.wlr_cursor_warp(self._ptr, input_device_ptr, x, y)
        elif warp_mode == WarpMode.LayoutClosest:
            lib.wlr_cursor_warp_closest(self._ptr, input_device_ptr, x, y)
            return True
        elif warp_mode == WarpMode.AbsoluteClosest:
            lib.wlr_cursor_warp_absolute(self._ptr, input_device_ptr, x, y)
            return True
        else:
            raise ValueError("Invalid warp mode")

    def absolute_to_layout_coords(
        self, input_device: InputDevice | None, x: float, y: float
    ) -> tuple[float, float]:
        """Convert absolute 0..1 coordinates to layout coordinates

        The `input_device` may be passed to respect device mapping constraints.
        If `input_device` is `None`, device mapping constraints will be
        ignored.
        """
        input_device_ptr = ptr_or_null(input_device)

        xy_ptr = ffi.new("double[2]")
        lib.wlr_cursor_absolute_to_layout_coords(
            self._ptr, input_device_ptr, x, y, xy_ptr, xy_ptr + 1
        )

        return xy_ptr[0], xy_ptr[1]

    def set_surface(self, surface: Surface | None, hotspot: tuple[int, int]) -> None:
        """Set the cursor surface

        The surface can be committed to update the cursor image. The surface
        position is subtracted from the hotspot. A None surface commit hides
        the cursor.
        """
        surface_ptr = ptr_or_null(surface)
        lib.wlr_cursor_set_surface(self._ptr, surface_ptr, hotspot[0], hotspot[1])

    def __enter__(self) -> Cursor:
        """Context manager to clean up the cursor"""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        """Clean up the cursor when exiting the context"""
        self.destroy()

    def map_to_output(self, output: Output | None) -> None:
        """
        Attaches this cursor to the given output, which must be among the outputs in the
        current output_layout for this cursor. This call is invalid for a cursor without
        an associated output layout.
        """
        output_ptr = ptr_or_null(output)
        lib.wlr_cursor_map_to_output(self._ptr, output_ptr)

    def map_input_to_output(
        self, input_device: InputDevice, output: Output | None
    ) -> None:
        """
        Maps all input from a specific input device to a given output. The input device
        must be attached to this cursor and the output must be among the outputs in the
        attached output layout.
        """
        output_ptr = ptr_or_null(output)
        lib.wlr_cursor_map_input_to_output(self._ptr, input_device._ptr, output_ptr)

    def set_xcursor(self, manager: XCursorManager, name: str) -> None:
        """
        Set the cursor image from an XCursor theme.

        The image will be loaded from the struct wlr_xcursor_manager.
        """
        lib.wlr_cursor_set_xcursor(self._ptr, manager._ptr, name.encode())



# Copyright (c) Matt Colligan 2023

from __future__ import annotations

from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

from pywayland.server import Signal

from wlroots import Ptr, ffi, lib

from .compositor import Surface

if TYPE_CHECKING:
    from pywayland.server import Display

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class XdgActivationV1(Ptr):
    def __init__(self, ptr) -> None:
        """An XDG activation manager: struct wlr_xdg_activation_v1."""
        self._ptr = ffi.cast("struct wlr_xdg_activation_v1 *", ptr)

        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))
        self.request_activate_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_activate),
            data_wrapper=XdgActivationV1RequestActivateEvent,
        )

    @classmethod
    def create(cls, display: Display) -> XdgActivationV1:
        """Create a `struct wlr_xdg_activation_v1` for the given display."""
        ptr = lib.wlr_xdg_activation_v1_create(display._ptr)
        return cls(ptr)


class XdgActivationV1RequestActivateEvent(Ptr):
    def __init__(self, ptr) -> None:
        """struct wlr_xdg_activation_v1_request_activate_event"""
        self._ptr = ffi.cast(
            "struct wlr_xdg_activation_v1_request_activate_event *", ptr
        )

    @property
    def surface(self) -> Surface:
        surface_ptr = self._ptr.surface
        _weakkeydict[surface_ptr] = self._ptr
        return Surface(surface_ptr)



# Copyright (c) 2019 Sean Vig

from .buffer import Buffer  # noqa: F401
from .compositor import (  # noqa: F401
    Compositor,
    SubCompositor,
    SubSurface,
    Surface,
    SurfaceState,
)
from .cursor import Cursor  # noqa: F401
from .data_control_v1 import DataControlManagerV1  # noqa: F401
from .data_device_manager import DataDeviceManager  # noqa: F401
from .export_dmabuf_v1 import ExportDmabufManagerV1  # noqa: F401
from .foreign_toplevel_management_v1 import ForeignToplevelManagerV1  # noqa: F401
from .fractional_scale_v1 import FractionalScaleManagerV1  # noqa: F401
from .gamma_control_v1 import GammaControlManagerV1  # noqa: F401
from .input_device import InputDevice  # noqa: F401
from .input_inhibit import InputInhibitManager  # noqa: F401
from .keyboard import Keyboard  # noqa: F401
from .layer_shell_v1 import LayerShellV1  # noqa: F401
from .matrix import Matrix  # noqa: F401
from .output import Output, OutputState  # noqa: F401
from .output_layout import OutputLayout, OutputLayoutOutput  # noqa: F401
from .pointer import (  # noqa: F401
    PointerAxisEvent,
    PointerButtonEvent,
    PointerMotionAbsoluteEvent,
    PointerMotionEvent,
)
from .pointer_constraints_v1 import (  # noqa: F401
    PointerConstraintsV1,
    PointerConstraintV1,
)
from .pointer_gestures_v1 import PointerGesturesV1  # noqa: F401
from .presentation_time import Presentation  # noqa: F401
from .primary_selection_v1 import PrimarySelectionV1DeviceManager  # noqa: F401
from .relative_pointer_manager_v1 import RelativePointerManagerV1  # noqa: F401
from .scene import (  # noqa: F401
    Scene,
    SceneBuffer,
    SceneNode,
    SceneNodeType,
    SceneOutput,
    SceneOutputLayout,
    SceneSurface,
    SceneTree,
)
from .screencopy_v1 import ScreencopyManagerV1  # noqa: F401
from .seat import Seat  # noqa: F401
from .texture import Texture  # noqa: F401
from .viewporter import Viewporter  # noqa: F401
from .virtual_keyboard_v1 import VirtualKeyboardManagerV1  # noqa: F401
from .virtual_pointer_v1 import VirtualPointerManagerV1  # noqa: F401
from .xcursor_manager import XCursorManager  # noqa: F401
from .xdg_decoration_v1 import XdgDecorationManagerV1  # noqa: F401
from .xdg_output_v1 import XdgOutputManagerV1  # noqa: F401
from .xdg_shell import XdgShell  # noqa: F401



# Copyright (c) Charbel Assaad 2023
from pywayland.server import Display

from wlroots import Ptr, lib

from .seat import Seat


class IdleNotifierV1(Ptr):
    def __init__(self, display: Display) -> None:
        self._ptr = lib.wlr_idle_notifier_v1_create(display._ptr)

    @property
    def inhibited(self) -> bool:
        return bool(self._ptr.inhibited)

    def set_inhibited(self, inhibited: bool) -> None:
        """
        Inhibit idle.

        Compositors should call this function when the idle state is disabled, e.g.
        because a visible client is using the idle-inhibit protocol.
        """
        lib.wlr_idle_notifier_v1_set_inhibited(self._ptr, inhibited)

    def notify_activity(self, seat: Seat) -> None:
        """
        Notify for user activity on a seat.

        Compositors should call this function whenever an input event is triggered
        on a seat.
        """
        lib.wlr_idle_notifier_v1_notify_activity(self._ptr, seat._ptr)



# Copyright (c) 2021 Graeme Holliday

import warnings

from pywayland.server import Client, Display, Signal

from wlroots import Ptr, ffi, lib


class InputInhibitManager(Ptr):
    def __init__(self, display: Display) -> None:
        """Creates a wlr_input_inhibit_manager"""
        warnings.warn(
            "Following the protocol deprecation, wlr/types/wlr_input_inhibitor.h is"
            " deprecated and will be removed in the next release.",
            DeprecationWarning,
            stacklevel=1,
        )

        self._ptr = lib.wlr_input_inhibit_manager_create(display._ptr)

        self.activate_event = Signal(ptr=ffi.addressof(self._ptr.events.activate))
        self.deactivate_event = Signal(ptr=ffi.addressof(self._ptr.events.deactivate))
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    def is_inactive(self) -> bool:
        return not self._ptr.active_inhibitor

    @property
    def active_client(self) -> Client:
        return Client(ptr=self._ptr.active_client)



# Copyright (c) Sean Vig 2018

from __future__ import annotations

import weakref
from typing import Any, Callable, TypeVar

from ._ffi import ffi, lib
from .version import version as _version

__wlroots_version__ = (
    f"{lib.WLR_VERSION_MAJOR}.{lib.WLR_VERSION_MINOR}.{lib.WLR_VERSION_MICRO}"
)

__version__ = _version

_weakkeydict: weakref.WeakKeyDictionary = weakref.WeakKeyDictionary()

T = TypeVar("T")


class Ptr:
    """Add equality checks for objects holding the same cdata

    Objects that reference the same cdata objects will be treated as equal.
    Note that these objects will still have a different hash such that they
    should not collide in a set or dictionary.
    """

    _ptr: ffi.CData

    def __eq__(self, other) -> bool:
        """Return true if the other object holds the same cdata"""
        return hasattr(other, "_ptr") and self._ptr == other._ptr

    def __hash__(self) -> int:
        """Use the hash from `object`, which is unique per object"""
        return super().__hash__()


class PtrHasData(Ptr):
    """
    Add methods to get and set the void *data member on the wrapped struct. The value
    stored can be of any Python type.
    """

    @property
    def data(self) -> Any | None:
        """Return any data that has been stored on the object"""
        if self._ptr.data == ffi.NULL:
            return None
        return ffi.from_handle(self._ptr.data)

    @data.setter
    def data(self, data: Any) -> None:
        """Store the given data on the current object"""
        if data is None:
            # Clear the data reference.
            if self.data in _weakkeydict:
                del _weakkeydict[self.data]
            self._ptr.data = ffi.NULL
            return

        # We adding a new data reference.
        if isinstance(data, ffi.CData):
            # We were already provided with a handle. The allows users of this code to
            # handle memory themselves.
            handle = data
        else:
            # We need to make a new handle. This keeps the data alive.
            handle = ffi.new_handle(data)
            _weakkeydict[data] = handle
        self._ptr.data = handle


def str_or_none(member: ffi.CData) -> str | None:
    """
    Helper function to check struct members for ffi.NULL, returning None, or a char
    array, returning a string.
    """
    if member:
        return ffi.string(member).decode(errors="backslashreplace")
    return None


def ptr_or_null(obj: Ptr | None) -> ffi.CData:
    """
    Helper function to check if the object is None, returning ffi.NULL,
    otherwise the _ptr attribute of the object
    """
    return obj._ptr if obj is not None else ffi.NULL


def instance_or_none(cls: Callable[[ffi.CData], T], ptr: ffi.CData) -> T | None:
    """
    A factory function which returns eiher an instance of T or None.

    The result depends on ``ptr`` if it is ffi.NULL, None will be returned,
    otherwise the result of T(ptr)
    """
    return cls(ptr) if ptr != ffi.NULL else None



from pywayland.server import Display

from wlroots import Ptr, lib


class SinglePixelBufferManagerV1(Ptr):
    """Global factory for single-pixel buffers.

    This protocol extension allows clients to create single-pixel buffers.

    Compositors supporting this protocol extension should also support the
    viewporter protocol extension (see wlr_types.viewporter.Viewporter).

    Clients may use viewporter to scale a single-pixel buffer to a desired size.
    """

    def __init__(self, display: Display) -> None:
        """Binds the manager to the provided display.

        :param display: The wayland display
        """
        self._ptr = lib.wlr_single_pixel_buffer_manager_v1_create(display._ptr)



# Copyright (c) 2024 Jeroen Wijenbergh

from pywayland.server import Display

from wlroots import PtrHasData, lib

from .compositor import Surface


class FractionalScaleManagerV1(PtrHasData):
    def __init__(self, display: Display, version: int = 1) -> None:
        """Create a wlr_fractional_scale_manager_v1"""
        self._ptr = lib.wlr_fractional_scale_manager_v1_create(display._ptr, version)


def notify_scale(surface: Surface, scale: float):
    lib.wlr_fractional_scale_v1_notify_scale(surface._ptr, scale)



# Copyright (c) Matt Colligan 2023

from __future__ import annotations

from typing import TYPE_CHECKING

from wlroots import Ptr, ffi, lib

if TYPE_CHECKING:
    from pywayland.server import Display

    from wlroots.backend import Backend


class Presentation(Ptr):
    def __init__(self, ptr) -> None:
        """A presentation time manager: struct wlr_presentation."""
        self._ptr = ffi.cast("struct wlr_presentation *", ptr)

    @classmethod
    def create(cls, display: Display, backend: Backend) -> Presentation:
        """Create a `struct wlr_xdg_activation_v1` for the given display."""
        ptr = lib.wlr_presentation_create(display._ptr, backend._ptr)
        return cls(ptr)



# Copyright (c) Matt Colligan 2022

from __future__ import annotations

import enum

from wlroots import Ptr, ffi, lib


class BufferDataPtrAccessFlag(enum.IntFlag):
    READ = lib.WLR_BUFFER_DATA_PTR_ACCESS_READ
    WRITE = lib.WLR_BUFFER_DATA_PTR_ACCESS_WRITE


class Buffer(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ptr

    def drop(self) -> None:
        """Destroys this wlr_texture."""
        lib.wlr_buffer_drop(self._ptr)

    def begin_data_ptr_access(
        self, flags: BufferDataPtrAccessFlag
    ) -> tuple[ffi.CData, int, int]:
        """Access a pointer to the underlying data"""
        data = ffi.new("void**")
        format_ptr = ffi.new("uint32_t *")
        stride_ptr = ffi.new("size_t *")

        if not lib.wlr_buffer_begin_data_ptr_access(
            self._ptr, flags, data, format_ptr, stride_ptr
        ):
            raise RuntimeError("Failed to get buffer data")

        return data[0], format_ptr[0], stride_ptr[0]

    def end_data_ptr_access(self) -> None:
        """End access to the underlying data"""
        lib.wlr_buffer_end_data_ptr_access(self._ptr)



# Copyright (c) Sean Vig 2022

from __future__ import annotations

import enum
from typing import TYPE_CHECKING, Callable, TypeVar

from pywayland.utils import wl_list_for_each

from wlroots import Ptr, PtrHasData, ffi, lib
from wlroots.util.region import PixmanRegion32
from wlroots.wlr_types import OutputLayoutOutput, Surface

if TYPE_CHECKING:
    from typing import Iterator

    from wlroots.util.box import Box
    from wlroots.util.clock import Timespec
    from wlroots.wlr_types import Buffer, Output, OutputLayout
    from wlroots.wlr_types.data_device_manager import DragIcon
    from wlroots.wlr_types.layer_shell_v1 import LayerSurfaceV1
    from wlroots.wlr_types.presentation_time import Presentation
    from wlroots.wlr_types.xdg_shell import XdgSurface


class SceneNodeType(enum.IntEnum):
    TREE = lib.WLR_SCENE_NODE_TREE
    RECT = lib.WLR_SCENE_NODE_RECT
    BUFFER = lib.WLR_SCENE_NODE_BUFFER


class Scene(Ptr):
    def __init__(self) -> None:
        """ "A root scene-graph node."""
        self._ptr = lib.wlr_scene_create()

    @property
    def tree(self) -> SceneTree:
        """The associated scene node."""
        ptr = ffi.addressof(self._ptr.tree)
        return SceneTree(ptr)

    def attach_output_layout(
        self, output_layout: OutputLayout
    ) -> SceneOutputLayout | None:
        """Attach an output layout to a scene."""
        ptr = lib.wlr_scene_attach_output_layout(self._ptr, output_layout._ptr)
        if ptr == ffi.NULL:
            return None
        return SceneOutputLayout(ptr)

    def set_presentation(self, presentation: Presentation) -> None:
        """
        Handle presentation feedback for all surfaces in the scene, assuming that scene
        outputs and the scene rendering functions are used.
        """
        lib.wlr_scene_set_presentation(self._ptr, presentation._ptr)

    def get_scene_output(self, output: Output) -> SceneOutput:
        """Get a scene-graph output from a wlr_output."""
        ptr = lib.wlr_scene_get_scene_output(self._ptr, output._ptr)
        return SceneOutput(ptr)

    @staticmethod
    def xdg_surface_create(parent: SceneTree, xdg_surface: XdgSurface) -> SceneTree:
        """Add a node displaying an xdg_surface and all of its sub-surfaces to the scene-graph.

        The origin of the returned scene-graph node will match the top-left
        corner of the xdg_surface window geometry.
        """
        ptr = lib.wlr_scene_xdg_surface_create(parent._ptr, xdg_surface._ptr)
        return SceneTree(ptr)

    @staticmethod
    def layer_surface_v1_create(
        parent: SceneTree, layer_surface: LayerSurfaceV1
    ) -> SceneLayerSurfaceV1:
        """
        Add a node displaying a layer_surface_v1 and all of its sub-surfaces to the
        scene-graph.

        The origin of the returned scene-graph node will match the top-left corner of
        the layer surface.
        """
        ptr = lib.wlr_scene_layer_surface_v1_create(parent._ptr, layer_surface._ptr)
        return SceneLayerSurfaceV1(ptr)


class SceneOutput(Ptr):
    def __init__(self, ptr) -> None:
        """A viewport for an output in the scene-graph"""
        self._ptr = ptr

    @classmethod
    def create(cls, scene: Scene, output: Output) -> SceneOutput:
        """
        Add a viewport for the specified output to the scene-graph.

        An output can only be added once to the scene-graph.
        """
        return cls(lib.wlr_scene_output_create(scene._ptr, output._ptr))

    def commit(self, options: SceneOutputStateOptions | None = None) -> bool:
        """Render and commit an output."""
        options_ptr = options._ptr if options is not None else ffi.NULL

        return lib.wlr_scene_output_commit(self._ptr, options_ptr)

    def destroy(self) -> None:
        """Destroy a scene-graph output."""
        lib.wlr_scene_output_destroy(self._ptr)

    def send_frame_done(self, timespec: Timespec) -> None:
        """Send frame done on scene output.

        Call send_frame_done() on all surfaces in the scene rendered by
        commit() for which scene_surface.primary_output matches the given
        scene_output.
        """
        lib.wlr_scene_output_send_frame_done(self._ptr, timespec._ptr)

    def set_position(self, lx: int, ly: int) -> None:
        """Set the output's position in the scene-graph."""
        lib.wlr_scene_output_set_position(self._ptr, lx, ly)


class SceneTree(PtrHasData):
    def __init__(self, ptr) -> None:
        """struct wlr_scene_tree"""
        self._ptr = ptr

    @property
    def node(self) -> SceneNode:
        """struct wlr_scene_tree"""
        ptr = ffi.addressof(self._ptr.node)
        return SceneNode(ptr)

    @classmethod
    def create(cls, parent: SceneTree) -> SceneTree:
        return SceneTree(lib.wlr_scene_tree_create(parent._ptr))

    @classmethod
    def subsurface_tree_create(cls, parent: SceneTree, surface: Surface) -> SceneTree:
        return SceneTree(
            lib.wlr_scene_subsurface_tree_create(parent._ptr, surface._ptr)
        )

    @classmethod
    def drag_icon_create(cls, parent: SceneTree, drag_icon: DragIcon) -> SceneTree:
        return SceneTree(lib.wlr_scene_drag_icon_create(parent._ptr, drag_icon._ptr))

    @property
    def children(self) -> Iterator[SceneNode]:
        for ptr in wl_list_for_each(
            "struct wlr_scene_node *",
            self._ptr.children,
            "link",
            ffi=ffi,
        ):
            yield SceneNode(ptr)


class SceneBuffer(Ptr):
    def __init__(self, ptr) -> None:
        """struct wlr_scene_buffer"""
        self._ptr = ptr

    @classmethod
    def from_node(cls, node: SceneNode) -> SceneBuffer | None:
        ptr = lib.wlr_scene_buffer_from_node(node._ptr)
        if ptr == ffi.NULL:
            return None
        return cls(ptr)

    @classmethod
    def create(cls, parent: SceneTree, buffer: Buffer) -> SceneBuffer | None:
        ptr = lib.wlr_scene_buffer_create(parent._ptr, buffer._ptr)
        if ptr == ffi.NULL:
            return None
        return cls(ptr)

    @property
    def node(self) -> SceneNode:
        ptr = ffi.addressof(self._ptr.node)
        return SceneNode(ptr)

    def set_buffer(self, buffer: Buffer | None) -> None:
        buffer_ptr = buffer._ptr if buffer else ffi.NULL
        lib.wlr_scene_buffer_set_buffer(self._ptr, buffer_ptr)

    def set_buffer_with_damage(
        self, buffer: Buffer | None, region: PixmanRegion32 | None = None
    ) -> None:
        buffer_ptr = buffer._ptr if buffer else ffi.NULL
        region_ptr = region._ptr if region else ffi.NULL
        lib.wlr_scene_buffer_set_buffer_with_damage(self._ptr, buffer_ptr, region_ptr)

    def set_opacity(self, opacity: float) -> None:
        """Sets the opacity of this buffer"""
        lib.wlr_scene_buffer_set_opacity(self._ptr, opacity)


T = TypeVar("T")
BufferCallback = Callable[[SceneBuffer, int, int, T], None]


@ffi.def_extern()
def buffer_iterator_callback(buffer_ptr, sx, sy, data_ptr):
    """Callback used to invoke the for_each_buffer method"""
    func, py_data = ffi.from_handle(data_ptr)
    buffer = SceneBuffer(buffer_ptr)
    func(buffer, sx, sy, py_data)


class SceneNode(PtrHasData):
    def __init__(self, ptr) -> None:
        """A node is an object in the scene."""
        self._ptr = ptr

    @property
    def type(self) -> SceneNodeType:
        return SceneNodeType(self._ptr.type)

    @property
    def parent(self) -> SceneTree | None:
        if self._ptr.parent == ffi.NULL:
            return None
        return SceneTree(self._ptr.parent)

    @property
    def x(self) -> int:
        return self._ptr.x

    @property
    def y(self) -> int:
        return self._ptr.y

    @property
    def enabled(self) -> bool:
        return self._ptr.enabled

    def destroy(self) -> None:
        """Immediately destroy the scene-graph node."""
        lib.wlr_scene_node_destroy(self._ptr)

    def set_enabled(self, *, enabled: bool = True) -> None:
        """
        Enable or disable this node. If a node is disabled, all of its children are
        implicitly disabled as well.
        """
        lib.wlr_scene_node_set_enabled(self._ptr, enabled)

    def set_position(self, x: int, y: int) -> None:
        """Set the position of the node relative to its parent."""
        lib.wlr_scene_node_set_position(self._ptr, x, y)

    def place_above(self, sibling: SceneNode) -> None:
        """Move the node right above the specified sibling."""
        lib.wlr_scene_node_place_above(self._ptr, sibling._ptr)

    def place_below(self, sibling: SceneNode) -> None:
        """Move the node right below the specified sibling."""
        lib.wlr_scene_node_place_below(self._ptr, sibling._ptr)

    def raise_to_top(self) -> None:
        """Move the node above all of its sibling nodes."""
        lib.wlr_scene_node_raise_to_top(self._ptr)

    def lower_to_bottom(self) -> None:
        """Move the node below all of its sibling nodes."""
        lib.wlr_scene_node_lower_to_bottom(self._ptr)

    def reparent(self, new_parent: SceneTree) -> None:
        """Move the node below all of its sibling nodes."""
        lib.wlr_scene_node_reparent(self._ptr, new_parent._ptr)

    def node_at(self, lx: float, ly: float) -> tuple[SceneNode, float, float] | None:
        """
        Find the topmost node in this scene-graph that contains the point at the given
        layout-local coordinates.

        (For surface nodes, this means accepting input events at that point.) Returns
        the node and coordinates relative to the returned node, or NULL if no node is
        found at that location.
        """
        nx = ffi.new("double*")
        ny = ffi.new("double*")
        node_ptr = lib.wlr_scene_node_at(self._ptr, lx, ly, nx, ny)
        if node_ptr == ffi.NULL:
            return None
        return SceneNode(node_ptr), nx[0], ny[0]

    def for_each_buffer(
        self, iterator: BufferCallback[T], data: T | None = None
    ) -> None:
        """
        Calls the iterator function for each sub-surface and popup of this surface
        """
        py_handle = (iterator, data)
        handle = ffi.new_handle(py_handle)
        lib.wlr_scene_node_for_each_buffer(
            self._ptr, lib.buffer_iterator_callback, handle
        )

    def subsurface_tree_set_clip(self, clip: Box | None):
        """
        Sets a cropping region for any subsurface trees that are children of this scene node.

        A None value will disable clipping
        """
        clip_ptr = ffi.NULL
        if clip is not None:
            clip_ptr = clip._ptr
        lib.wlr_scene_subsurface_tree_set_clip(self._ptr, clip_ptr)


class SceneSurface(Ptr):
    def __init__(self, ptr) -> None:
        """struct wlr_scene_surface"""
        self._ptr = ptr

    @classmethod
    def from_buffer(cls, buffer: SceneBuffer) -> SceneSurface | None:
        ptr = lib.wlr_scene_surface_try_from_buffer(buffer._ptr)
        if ptr == ffi.NULL:
            return None
        return cls(ptr)

    @property
    def surface(self) -> Surface:
        return Surface(self._ptr.surface)


class SceneRect(Ptr):
    def __init__(
        self, parent: SceneTree, width: int, height: int, color: ffi.CData
    ) -> None:
        """A scene-graph node displaying a solid-colored rectangle"""
        self._ptr = lib.wlr_scene_rect_create(parent._ptr, width, height, color)

    @property
    def node(self) -> SceneNode:
        """struct wlr_scene_tree"""
        ptr = ffi.addressof(self._ptr.node)
        return SceneNode(ptr)

    def set_size(self, width: int, height: int) -> None:
        """Change the width and height of an existing rectangle node."""
        lib.wlr_scene_rect_set_size(self._ptr, width, height)

    def set_color(self, color: ffi.CData) -> None:
        """Change the color of an existing rectangle node."""
        lib.wlr_scene_rect_set_color(self._ptr, color)


class SceneLayerSurfaceV1(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ptr

    @property
    def tree(self) -> SceneTree:
        """struct wlr_scene_tree"""
        return SceneTree(self._ptr.tree)

    def configure(self, full_area: Box, usable_area: Box) -> None:
        """
        Configure a layer_surface_v1, position its scene node in accordance to its
        current state, and update the remaining usable area.

        full_area represents the entire area that may be used by the layer surface if
        its exclusive_zone is -1, and is usually the output dimensions. usable_area
        represents what remains of full_area that can be used if exclusive_zone is >= 0.
        usable_area is updated if the surface has a positive exclusive_zone, so that it
        can be used for the next layer surface.
        """
        lib.wlr_scene_layer_surface_v1_configure(
            self._ptr, full_area._ptr, usable_area._ptr
        )


class SceneOutputLayout(Ptr):
    def __init__(self, ptr) -> None:
        """A `struct wlr_scene_output_layout_scene`"""
        self._ptr = ptr

    def add_output(
        self, output_layout_output: OutputLayoutOutput, scene_output: SceneOutput
    ) -> None:
        """Add an output to the scene output layout."""
        lib.wlr_scene_output_layout_add_output(
            self._ptr, output_layout_output._ptr, scene_output._ptr
        )


class SceneOutputStateOptions(Ptr):
    def __init__(self, ptr) -> None:
        """A `struct wlr_scene_output_state_options`."""
        self._ptr = ptr



from __future__ import annotations

import enum
from weakref import WeakKeyDictionary

from pywayland.server import Signal

from wlroots import Ptr, PtrHasData, ffi, lib

from .input_device import InputDevice

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


@enum.unique
class SwitchType(enum.IntEnum):
    LID = lib.WLR_SWITCH_TYPE_LID
    TABLET_MODE = lib.WLR_SWITCH_TYPE_TABLET_MODE


@enum.unique
class SwitchState(enum.IntEnum):
    STATE_OFF = lib.WLR_SWITCH_STATE_OFF
    STATE_ON = lib.WLR_SWITCH_STATE_ON


class Switch(PtrHasData):
    """
    A switch input device.

    Typically a switch input device can indicate whether a laptop lid is opened
    or closed, or whether tablet mode is enabled.

    See https://wayland.freedesktop.org/libinput/doc/latest/switches.html
    """

    def __init__(self, ptr) -> None:
        self._ptr = ptr
        self.toggle_event = Signal(
            ptr=ffi.addressof(self._ptr.events.toggle), data_wrapper=SwitchToggleEvent
        )

    @staticmethod
    def from_input_device(input_device: InputDevice) -> Switch:
        """
        Get a Switch from an InputDevice.

        Asserts that the input device is a switch.
        """
        return Switch(lib.wlr_switch_from_input_device(input_device._ptr))

    @property
    def base(self) -> InputDevice:
        device_ptr = ffi.addressof(self._ptr.base)
        _weakkeydict[device_ptr] = self._ptr
        return InputDevice(device_ptr)


class SwitchToggleEvent(Ptr):

    def __init__(self, ptr):
        self._ptr = ptr

    @property
    def time_msec(self) -> int:
        return self._ptr.time_msec

    @property
    def switch_type(self) -> SwitchType:
        return SwitchType(self._ptr.switch_type)

    @property
    def switch_state(self) -> SwitchState:
        return SwitchState(self._ptr.switch_state)



# Copyright (c) Sean Vig 2019

from __future__ import annotations

from pywayland.server import Signal

from wlroots import Ptr, ffi, lib
from wlroots.util.box import Box

from .output import Output


class OutputLayout(Ptr):
    def __init__(self) -> None:
        """Creates an output layout to work with a layout of screens

        Creates a output layout, which can be used to describing outputs in
        physical space relative to one another, and perform various useful
        operations on that state.
        """
        ptr = lib.wlr_output_layout_create()
        self._ptr = ffi.gc(ptr, lib.wlr_output_layout_destroy)

        self.add_event = Signal(ptr=ffi.addressof(ptr.events.add))
        self.change_event = Signal(ptr=ffi.addressof(ptr.events.change))
        self.destroy_event = Signal(ptr=ffi.addressof(ptr.events.destroy))

    def destroy(self) -> None:
        """Destroy the current output layout"""
        if self._ptr is not None:
            ffi.release(self._ptr)
            self._ptr = None

    def add_auto(self, output: Output) -> OutputLayoutOutput | None:
        """
        Add the output to the layout as automatically configured. This will place the
        output in a sensible location in the layout. The coordinates of the output in
        the layout will be adjusted dynamically when the layout changes. If the output
        is already a part of the layout, it will become automatically configured.

        Returns true on success, false on a memory allocation error.
        """
        ptr = lib.wlr_output_layout_add_auto(self._ptr, output._ptr)
        if ptr == ffi.NULL:
            return None
        return OutputLayoutOutput(ptr)

    def output_coords(self, output: Output) -> tuple[float, float]:
        """Determine coordinates of the output in the layout

        Given x and y in layout coordinates, adjusts them to local output
        coordinates relative to the given reference output.
        """
        ox = ffi.new("double *")
        oy = ffi.new("double *")
        lib.wlr_output_layout_output_coords(self._ptr, output._ptr, ox, oy)

        return ox[0], oy[0]

    def __enter__(self) -> OutputLayout:
        """Use the output layout in a context manager"""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        """Clean up the output layout when exiting the context"""
        self.destroy()

    def output_at(self, x: float, y: float) -> Output | None:
        """
        Get the output at the specified layout coordinates. Returns None if no output
        matches the coordinates.
        """
        output_ptr = lib.wlr_output_layout_output_at(self._ptr, x, y)
        if output_ptr == ffi.NULL:
            return None
        return Output(output_ptr)

    def add(self, output: Output, lx: int, ly: int) -> OutputLayoutOutput | None:
        """
        Add the output to the layout at the specified coordinates. If the output is
        already a part of the output layout, it will become manually configured and will
        be moved to the specified coordinates.
        """
        ptr = lib.wlr_output_layout_add(self._ptr, output._ptr, lx, ly)
        if ptr == ffi.NULL:
            return None
        return OutputLayoutOutput(ptr)

    def remove(self, output: Output) -> None:
        """Remove an output from the layout."""
        lib.wlr_output_layout_remove(self._ptr, output._ptr)

    def get_box(
        self, reference: Output | None = None, dest_box: Box | None = None
    ) -> Box:
        """
        Get the box of the layout for the given reference output in layout
        coordinates. If `reference` is None, the box will be for the extents of the
        entire layout. If the output isn't in the layout, the box will be empty.
        """
        if reference:
            reference_ptr = reference._ptr
        else:
            reference_ptr = ffi.NULL

        if not dest_box:
            dest_box = Box(ptr=ffi.new("struct wlr_box *"))

        lib.wlr_output_layout_get_box(self._ptr, reference_ptr, dest_box._ptr)
        return dest_box

    def closest_point(
        self, lx: float, ly: float, reference: Output | None = None
    ) -> tuple[float, float]:
        """
        Get the closest point on this layout from the given point from the reference
        output. If reference is NULL, gets the closest point from the entire layout.
        """
        if reference:
            reference_ptr = reference._ptr
        else:
            reference_ptr = ffi.NULL

        dest_lx = ffi.new("double *")
        dest_ly = ffi.new("double *")
        lib.wlr_output_layout_closest_point(
            self._ptr, reference_ptr, lx, ly, dest_lx, dest_ly
        )
        return dest_lx[0], dest_ly[0]


class OutputLayoutOutput(Ptr):
    def __init__(self, ptr) -> None:
        """A `struct wlr_output_layout_output`"""
        self._ptr = ptr



# Copyright (c) Matt Colligan 2022

from __future__ import annotations

from typing import TYPE_CHECKING

from pywayland.server import Signal

from wlroots import PtrHasData, ffi, lib

if TYPE_CHECKING:
    from pywayland.server import Display

    from .seat import Seat


class PointerGesturesV1(PtrHasData):
    def __init__(self, display: Display) -> None:
        """Manager to relay pointer gestures to clients.

        :param display:
            The display to relay gestures for.
        """
        self._ptr = lib.wlr_pointer_gestures_v1_create(display._ptr)

        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    def send_swipe_begin(self, seat: Seat, time_msec: int, fingers: int) -> None:
        lib.wlr_pointer_gestures_v1_send_swipe_begin(
            self._ptr,
            seat._ptr,
            time_msec,
            fingers,
        )

    def send_swipe_update(
        self, seat: Seat, time_msec: int, dx: float, dy: float
    ) -> None:
        lib.wlr_pointer_gestures_v1_send_swipe_update(
            self._ptr,
            seat._ptr,
            time_msec,
            dx,
            dy,
        )

    def send_swipe_end(self, seat: Seat, time_msec: int, cancelled: bool) -> None:
        lib.wlr_pointer_gestures_v1_send_swipe_end(
            self._ptr,
            seat._ptr,
            time_msec,
            cancelled,
        )

    def send_pinch_begin(self, seat: Seat, time_msec: int, fingers: int) -> None:
        lib.wlr_pointer_gestures_v1_send_pinch_begin(
            self._ptr,
            seat._ptr,
            time_msec,
            fingers,
        )

    def send_pinch_update(
        self,
        seat: Seat,
        time_msec: int,
        dx: float,
        dy: float,
        scale: float,
        rotation: float,
    ) -> None:
        lib.wlr_pointer_gestures_v1_send_pinch_update(
            self._ptr,
            seat._ptr,
            time_msec,
            dx,
            dy,
            scale,
            rotation,
        )

    def send_pinch_end(self, seat: Seat, time_msec: int, cancelled: bool) -> None:
        lib.wlr_pointer_gestures_v1_send_pinch_end(
            self._ptr,
            seat._ptr,
            time_msec,
            cancelled,
        )

    def send_hold_begin(self, seat: Seat, time_msec: int, fingers: int) -> None:
        lib.wlr_pointer_gestures_v1_send_hold_begin(
            self._ptr,
            seat._ptr,
            time_msec,
            fingers,
        )

    def send_hold_end(self, seat: Seat, time_msec: int, cancelled: bool) -> None:
        lib.wlr_pointer_gestures_v1_send_hold_end(
            self._ptr,
            seat._ptr,
            time_msec,
            cancelled,
        )



"""
Support for the ext-session-lock-v1 protocol.

Secure session locking with arbitrary graphics

This protocol allows for a privileged Wayland client to lock the session and
display arbitrary graphics while the session is locked.

The compositor may choose to restrict this protocol to a special client
launched by the compositor itself or expose it to all privileged clients,
this is compositor policy.

The client is responsible for performing authentication and informing the
compositor when the session should be unlocked. If the client dies while the
session is locked the session remains locked, possibly permanently depending
on compositor policy.
"""

from __future__ import annotations

from weakref import WeakKeyDictionary

from pywayland.server import Display, Signal

from wlroots import PtrHasData, ffi, lib
from wlroots.wlr_types.output import Output

from .compositor import Surface

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class SessionLockManagerV1(PtrHasData):

    def __init__(self, display: Display) -> None:
        self._ptr = lib.wlr_session_lock_manager_v1_create(display._ptr)
        self.new_lock_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_lock), data_wrapper=SessionLockV1
        )
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))


class SessionLockV1(PtrHasData):

    def __init__(self, ptr):
        self._ptr = ffi.cast("struct wlr_session_lock_v1 *", ptr)
        self.new_surface_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_surface),
            data_wrapper=SessionLockSurfaceV1,
        )
        self.unlock_event = Signal(ptr=ffi.addressof(self._ptr.events.unlock))
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    def send_locked(self):
        lib.wlr_session_lock_v1_send_locked(self._ptr)

    def destroy(self):
        lib.wlr_session_lock_v1_destroy(self._ptr)


class SessionLockSurfaceV1(PtrHasData):
    """
    A surface displayed while the session is locked
    """

    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_session_lock_surface_v1 *", ptr)
        self.map_event = Signal(ptr=ffi.addressof(self._ptr.events.map))
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    @property
    def output(self) -> Output:
        output_ptr = self._ptr.output
        _weakkeydict[output_ptr] = self._ptr
        return Output(output_ptr)

    @property
    def surface(self) -> Surface:
        surface_ptr = self._ptr.surface
        _weakkeydict[surface_ptr] = self._ptr
        return Surface(surface_ptr)

    @property
    def configured(self) -> bool:
        return self._ptr.configured

    @property
    def mapped(self) -> bool:
        return self._ptr.mapped

    def configure(self, width: int, height: int) -> int:
        """
        Configures the width and height of the surface.

        Returns the configure serial.
        """
        return lib.wlr_session_lock_surface_v1_configure(self._ptr, width, height)

    @staticmethod
    def try_from_surface(surface: Surface) -> SessionLockSurfaceV1 | None:
        """
        Get a SessionLockSurfaceV1 from a surface.

        Asserts that the surface has the session lock surface role.
        May return None even if the surface has the session lock surface role if the
        corresponding session lock surface has been destroyed.
        """
        surface_ptr = lib.wlr_session_lock_surface_v1_try_from_wlr_surface(surface._ptr)
        return SessionLockSurfaceV1(surface_ptr) if surface_ptr != ffi.NULL else None



# Copyright (c) 2021 Matt Colligan

from weakref import WeakKeyDictionary

from pywayland.server import Display, Signal

from wlroots import Ptr, ffi, lib
from wlroots.wlr_types.keyboard import Keyboard

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class VirtualKeyboardManagerV1(Ptr):
    def __init__(self, display: Display) -> None:
        """A wlr_virtual_keyboard_manager_v1 instance."""
        self._ptr = lib.wlr_virtual_keyboard_manager_v1_create(display._ptr)

        self.new_virtual_keyboard_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_virtual_keyboard),
            data_wrapper=VirtualKeyboardV1,
        )
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))


class VirtualKeyboardV1(Ptr):
    def __init__(self, ptr) -> None:
        """A wlr_virtual_keyboard_v1 instance."""
        self._ptr = ffi.cast("struct wlr_virtual_keyboard_v1 *", ptr)

    @property
    def keyboard(self) -> Keyboard:
        keyboard_ptr = ffi.addressof(self._ptr.keyboard)
        _weakkeydict[keyboard_ptr] = self._ptr
        return Keyboard(keyboard_ptr)

    @property
    def has_keymap(self) -> bool:
        return self._ptr.has_keymap



# Copyright (c) Matt Colligan 2021
from __future__ import annotations

from weakref import WeakKeyDictionary

from wlroots import Ptr, PtrHasData, ffi, lib, str_or_none

from .input_device import InputDevice

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class Touch(PtrHasData):
    def __init__(self, ptr) -> None:
        self._ptr = ptr

    @staticmethod
    def from_input_device(input_device: InputDevice) -> Touch:
        return Touch(lib.wlr_touch_from_input_device(input_device._ptr))

    @property
    def base(self) -> InputDevice:
        device_ptr = ffi.addressof(self._ptr.base)
        _weakkeydict[device_ptr] = self._ptr
        return InputDevice(device_ptr)

    @property
    def output_name(self) -> str | None:
        """The name of any associated output"""
        return str_or_none(self._ptr.output_name)

    @property
    def width_mm(self) -> float:
        return self._ptr.width_mm

    @property
    def height_mm(self) -> float:
        return self._ptr.height_mm


class _TouchEvent(Ptr):
    @property
    def touch(self) -> Touch:
        """The touch device associated with the event"""
        return Touch(self._ptr.touch)

    @property
    def time_msec(self) -> int:
        return self._ptr.time_msec

    @property
    def touch_id(self) -> int:
        return self._ptr.touch_id


class TouchDownEvent(_TouchEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_touch_down_event *", ptr)

    @property
    def x(self) -> float:
        return self._ptr.x

    @property
    def y(self) -> float:
        return self._ptr.y


class TouchUpEvent(_TouchEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_touch_up_event *", ptr)


class TouchMotionEvent(_TouchEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_touch_motion_event *", ptr)

    @property
    def x(self) -> float:
        return self._ptr.x

    @property
    def y(self) -> float:
        return self._ptr.y


class TouchCancelEvent(_TouchEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_touch_cancel_event *", ptr)



# Copyright (c) Sean Vig 2019

from __future__ import annotations

from typing import TYPE_CHECKING

from wlroots import Ptr, ffi, lib

if TYPE_CHECKING:
    from typing import Iterator


class XCursorManager(Ptr):
    def __init__(self, theme: str | None, size: int = 24, scale: float = 1.0):
        """Creates a new XCursor manager using the theme and size

        and ensures an xcursor with scale is loaded
        """
        theme_ptr = ffi.NULL
        if theme is not None:
            theme_ptr = theme.encode()
        ptr = lib.wlr_xcursor_manager_create(theme_ptr, size)
        self._ptr = ffi.gc(ptr, lib.wlr_xcursor_manager_destroy)

        lib.wlr_xcursor_manager_load(self._ptr, scale)

    def destroy(self):
        """Destroy the x cursor manager"""
        if self._ptr is not None:
            ffi.release(self._ptr)
            self._ptr = None

    def get_xcursor(self, name: str, scale: float = 1) -> XCursor | None:
        """
        Retrieves a wlr_xcursor reference for the given cursor name at the given scale
        factor, or NULL if this wlr_xcursor_manager has not loaded a cursor theme at the
        requested scale.
        """
        ptr = lib.wlr_xcursor_manager_get_xcursor(self._ptr, name.encode(), scale)
        if ptr == ffi.NULL:
            return None
        return XCursor(ptr)

    def __enter__(self):
        """Setup X cursor manager in a context manager"""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Clean-up the X cursor manager on contex manager exit"""
        self.destroy()

    def load(self, scale: float) -> bool:
        """
        Ensures an xcursor theme at the given scale factor is loaded in the manager.
        """
        return bool(lib.wlr_xcursor_manager_load(self._ptr, scale))


class XCursor(Ptr):
    """struct wlr_xcursor"""

    def __init__(self, ptr):
        self._ptr = ptr

    @property
    def images(self) -> Iterator[XCursorImage]:
        for i in range(self._ptr.image_count):
            yield XCursorImage(self._ptr.images[0][i])


class XCursorImage(Ptr):
    """struct wlr_xcursor_image"""

    def __init__(self, ptr):
        self._ptr = ptr



# Copyright (c) 2019 Sean Vig

from __future__ import annotations

from typing import TYPE_CHECKING, Final
from weakref import WeakKeyDictionary

from pywayland.protocol.wayland import WlOutput
from pywayland.server import Display, Signal

from wlroots import Ptr, PtrHasData, ffi, lib
from wlroots.util.clock import Timespec

from .texture import Texture

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()

if TYPE_CHECKING:
    from wlroots.renderer import Renderer


_MAX_COMPOSITOR_VERSION: Final = 5


class Compositor(Ptr):
    def __init__(
        self, display: Display, version: int, renderer: Renderer | None = None
    ) -> None:
        """A compositor for clients to be able to allocate surfaces

        :param display:
            The Wayland server display to attach to the compositor.
        :param version:
            The version of the wlr_compositor interface to use.
        :param renderer:
            The wlroots renderer to attach the compositor to.
        """
        if not 0 < version <= _MAX_COMPOSITOR_VERSION:
            raise ValueError(
                f"Invalid compositor version, should be a value between 1 (inclusive) and {_MAX_COMPOSITOR_VERSION} (inclusive), got: {version}"
            )
        if renderer is None:
            self._ptr = lib.wlr_compositor_create(display._ptr, version, ffi.NULL)
        else:
            self._ptr = lib.wlr_compositor_create(display._ptr, version, renderer._ptr)


class SubCompositor(Ptr):
    def __init__(self, display: Display) -> None:
        self._ptr = lib.wlr_subcompositor_create(display._ptr)


class Surface(PtrHasData):
    def __init__(self, ptr) -> None:
        """Create a wlroots Surface

        :param ptr:
            The cdata for the given surface
        """
        self._ptr = ptr

        self.precommit_event = Signal(
            ptr=ffi.addressof(self._ptr.events.precommit),
            data_wrapper=SurfaceState,
        )
        self.commit_event = Signal(ptr=ffi.addressof(self._ptr.events.commit))
        self.map_event = Signal(ptr=ffi.addressof(self._ptr.events.map))
        self.unmap_event = Signal(ptr=ffi.addressof(self._ptr.events.unmap))
        self.new_subsurface_event = Signal(
            ptr=ffi.addressof(self._ptr.events.new_subsurface),
            data_wrapper=SubSurface,
        )
        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    @property
    def current(self) -> SurfaceState:
        """The current commited surface state"""
        current_ptr = self._ptr.current
        _weakkeydict[current_ptr] = self._ptr
        return SurfaceState(current_ptr)

    @property
    def previous(self) -> SurfaceState:
        """The state of the previous commit"""
        previous_ptr = self._ptr.previous
        _weakkeydict[previous_ptr] = self._ptr
        return SurfaceState(previous_ptr)

    def get_texture(self):
        """Get the texture of the buffer currently attached to this surface

        Returns None if no buffer is currently attached or if something went
        wrong with uploading the buffer.
        """
        texture_ptr = lib.wlr_surface_get_texture(self._ptr)
        if texture_ptr == ffi.NULL:
            return None

        return Texture(texture_ptr)

    def send_frame_done(self, when: Timespec) -> None:
        """Send a frame done event to the surface"""
        lib.wlr_surface_send_frame_done(self._ptr, when._ptr)


class SurfaceState(Ptr):
    def __init__(self, ptr):
        """The state of a given surface

        :param ptr:
            The cdata of the given surface state.
        """
        self._ptr = ptr

    @property
    def transform(self) -> WlOutput.transform:
        """Get the transform for the selected surface"""
        return WlOutput.transform(self._ptr.transform)

    @property
    def width(self) -> int:
        """In surface local width"""
        return self._ptr.width

    @property
    def height(self) -> int:
        """In surface local height"""
        return self._ptr.height


class SubSurface(PtrHasData):
    def __init__(self, ptr):
        """A wlroots subsurface

        :param ptr:
            The cdata for the given subsurface
        """
        self._ptr = ffi.cast("struct wlr_subsurface *", ptr)

        self.destroy_event = Signal(ptr=ffi.addressof(self._ptr.events.destroy))

    @property
    def surface(self) -> Surface:
        """Get the wlr_surface underlying this subsurface"""
        return Surface(self._ptr.surface)



# Copyright (c) 2019 Sean Vig

from __future__ import annotations

from typing import Iterator
from weakref import WeakKeyDictionary

from pywayland.protocol.wayland import WlSeat
from pywayland.server import Display, Signal
from pywayland.utils import wl_list_for_each

from wlroots import Ptr, PtrHasData, ffi, instance_or_none, lib, ptr_or_null

from .compositor import Surface
from .data_device_manager import Drag
from .input_device import ButtonState
from .keyboard import Keyboard, KeyboardKeyEvent, KeyboardModifiers
from .pointer import AxisOrientation, AxisSource

_weakkeydict: WeakKeyDictionary = WeakKeyDictionary()


class KeyboardGrab(Ptr):
    def __init__(self, seat: Seat) -> None:
        """Setup the keyboard grab"""
        self._ptr = ffi.new("struct wlr_seat_keyboard_grab *")
        self._seat = seat

    def __enter__(self) -> KeyboardGrab:
        """State the keyboard grab"""
        lib.wlr_seat_keyboard_start_grab(self._seat._ptr, self._ptr)
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        """End the grab of the keyboard of this seat"""
        lib.wlr_seat_keyboard_end_grab(self._seat._ptr)


class Seat(PtrHasData):
    def __init__(self, display: Display, name: str) -> None:
        """Allocates a new seat and adds a seat global to the display

        :param display:
            The Wayland server display to attach the seat to
        :param name:
            The name of the seat to create
        """
        ptr = lib.wlr_seat_create(display._ptr, name.encode())
        self._ptr = ffi.gc(ptr, lib.wlr_seat_destroy)

        self.pointer_grab_begin_event = Signal(
            ptr=ffi.addressof(self._ptr.events.pointer_grab_begin)
        )
        self.pointer_grab_end_event = Signal(
            ptr=ffi.addressof(self._ptr.events.pointer_grab_end)
        )

        self.keyboard_grab_begin_event = Signal(
            ptr=ffi.addressof(self._ptr.events.keyboard_grab_begin)
        )
        self.keyboard_grab_end_event = Signal(
            ptr=ffi.addressof(self._ptr.events.keyboard_grab_end)
        )

        self.touch_grab_begin_event = Signal(
            ptr=ffi.addressof(self._ptr.events.touch_grab_begin)
        )
        self.touch_grab_end_event = Signal(
            ptr=ffi.addressof(self._ptr.events.touch_grab_end)
        )

        self.request_set_cursor_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_set_cursor),
            data_wrapper=PointerRequestSetCursorEvent,
        )

        # Called when an application _wants_ to set the selection
        self.request_set_selection_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_set_selection),
            data_wrapper=RequestSetSelectionEvent,
        )
        # Called after the data source is set for the selection
        self.set_selection_event = Signal(
            ptr=ffi.addressof(self._ptr.events.set_selection)
        )

        # Called when an application _wants_ to set the primary selection (user selects some data)
        self.request_set_primary_selection_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_set_primary_selection),
            data_wrapper=RequestSetPrimarySelectionEvent,
        )
        # Called after the primary selection source object is set
        self.set_primary_selection_event = Signal(
            ptr=ffi.addressof(self._ptr.events.set_primary_selection)
        )

        self.request_start_drag_event = Signal(
            ptr=ffi.addressof(self._ptr.events.request_start_drag),
            data_wrapper=RequestStartDragEvent,
        )
        self.start_drag_event = Signal(
            ptr=ffi.addressof(self._ptr.events.start_drag),
            data_wrapper=Drag,
        )

    @property
    def pointer_state(self) -> SeatPointerState:
        """The pointer state associated with the seat"""
        pointer_state_ptr = ffi.addressof(self._ptr.pointer_state)
        _weakkeydict[pointer_state_ptr] = self._ptr
        return SeatPointerState(pointer_state_ptr)

    @property
    def keyboard_state(self) -> SeatKeyboardState:
        """The keyboard state associated with the seat"""
        keyboard_state_ptr = ffi.addressof(self._ptr.keyboard_state)
        _weakkeydict[keyboard_state_ptr] = self._ptr
        return SeatKeyboardState(keyboard_state_ptr)

    def get_keyboard(self) -> Keyboard | None:
        """Get the active keyboard for the seat."""
        return instance_or_none(Keyboard, lib.wlr_seat_get_keyboard(self._ptr))

    def set_keyboard(self, keyboard: Keyboard | None) -> None:
        """Set this keyboard as the active keyboard for the seat"""
        lib.wlr_seat_set_keyboard(self._ptr, ptr_or_null(keyboard))

    def destroy(self) -> None:
        """Clean up the seat"""
        if self._ptr is not None:
            ffi.release(self._ptr)
            self._ptr = None

    @property
    def destroyed(self) -> bool:
        """Whether this seat has been destroyed."""
        return self._ptr is None

    def set_capabilities(self, capabilities: WlSeat.capability) -> None:
        """Updates the capabilities available on this seat

        Will automatically send them to all clients.

        :param capabilities:
            The Wayland seat capabilities to set on the seat.
        """
        lib.wlr_seat_set_capabilities(self._ptr, capabilities)

    def set_name(self, name: str) -> None:
        """Updates the name of this seat

        Will automatically send it to all clients.
        """
        lib.wlr_seat_set_name(self._ptr, name.encode())

    def pointer_surface_has_focus(self, surface: Surface) -> bool:
        """Whether or not the surface has pointer focus"""
        return lib.wlr_seat_pointer_surface_has_focus(self._ptr, surface._ptr)

    def pointer_clear_focus(self) -> None:
        """Clear the focused surface for the pointer and leave all entered surfaces.

        This function does not respect pointer grabs: you probably want
        `pointer_notify_clear_focus()` instead.
        """
        return lib.wlr_seat_pointer_clear_focus(self._ptr)

    def pointer_notify_clear_focus(self) -> None:
        """Notify the seat of a pointer leave event to the currently-focused surface.

        Defers to any grab of the pointer.
        """
        return lib.wlr_seat_pointer_notify_clear_focus(self._ptr)

    def pointer_notify_enter(
        self, surface: Surface, surface_x: float, surface_y: float
    ) -> None:
        """Notify the seat of a pointer enter event to the given surface

        Notify the seat of a pointer enter event to the given surface and
        request it to be the focused surface for the pointer. Pass
        surface-local coordinates where the enter occurred.
        """
        lib.wlr_seat_pointer_notify_enter(self._ptr, surface._ptr, surface_x, surface_y)

    def pointer_notify_motion(
        self, time_msec: int, surface_x: float, surface_y: float
    ) -> None:
        """Notify the seat of motion over the given surface

        Pass surface-local coordinates where the pointer motion occurred.
        """
        lib.wlr_seat_pointer_notify_motion(self._ptr, time_msec, surface_x, surface_y)

    def pointer_notify_button(
        self, time_msec: int, button: int, button_state: ButtonState
    ) -> int:
        """Notify the seat that a button has been pressed

        Returns the serial of the button press or zero if no button press was
        sent.
        """
        return lib.wlr_seat_pointer_notify_button(
            self._ptr, time_msec, button, button_state.value
        )

    def pointer_notify_axis(
        self,
        time_msec: int,
        orientation: AxisOrientation,
        value: float,
        value_discrete: int,
        source: AxisSource,
    ) -> None:
        """Notify the seat of an axis event"""
        lib.wlr_seat_pointer_notify_axis(
            self._ptr, time_msec, orientation.value, value, value_discrete, source.value
        )

    def pointer_notify_frame(self) -> None:
        """Notify the seat of a frame event

        Frame events are sent to end a group of events that logically belong
        together. Motion, button and axis events should all be followed by a
        frame event.
        """
        lib.wlr_seat_pointer_notify_frame(self._ptr)

    def pointer_has_grab(self) -> bool:
        """Whether or not the pointer has a grab other than the default grab"""
        return lib.wlr_seat_pointer_has_grab(self._ptr)

    def grab(self) -> KeyboardGrab:
        """Start a grab of the keyboard of this seat"""
        return KeyboardGrab(self)

    def keyboard_notify_key(self, key_event: KeyboardKeyEvent) -> None:
        """Notify the seat that a key has been pressed on the keyboard

        Defers to any keyboard grabs.
        """
        lib.wlr_seat_keyboard_notify_key(
            self._ptr, key_event.time_msec, key_event.keycode, key_event.state
        )

    def keyboard_notify_modifiers(self, modifiers: KeyboardModifiers) -> None:
        """Notify the seat that the modifiers for the keyboard have changed

        Defers to any keyboard grabs.
        """
        lib.wlr_seat_keyboard_notify_modifiers(self._ptr, modifiers._ptr)

    def keyboard_notify_enter(self, surface: Surface, keyboard: Keyboard) -> None:
        """Notify the seat that the keyboard focus has changed

        Notify the seat that the keyboard focus has changed and request it to
        be the focused surface for this keyboard. Defers to any current grab of
        the seat's keyboard.
        """
        lib.wlr_seat_keyboard_notify_enter(
            self._ptr,
            surface._ptr,
            keyboard.keycodes,
            keyboard.num_keycodes,
            keyboard.modifiers._ptr,
        )

    def keyboard_clear_focus(self) -> None:
        """Clear the focused surface for the keyboard and leave all entered surfaces"""
        lib.wlr_seat_keyboard_clear_focus(self._ptr)

    def keyboard_has_grab(self) -> bool:
        """Whether or not the keyboard has a grab other than the default grab"""
        return lib.wlr_seat_keyboard_has_grab(self._ptr)

    def touch_send_down(
        self,
        surface: Surface,
        time_msec: int,
        touch_id: int,
        surface_x: float,
        surface_y: float,
    ) -> int:
        """
        Send a touch down event to the client of the given surface. All future touch
        events for this point will go to this surface. If the touch down is valid,
        this will add a new touch point with the given `touch_id`. The touch down may
        not be valid if the surface seat client does not accept touch input.
        Coordinates are surface-local. This function does not respect touch grabs:
        you probably want `touch_notify_down()` instead.
        """
        return lib.wlr_seat_touch_send_down(
            self._ptr, surface._ptr, time_msec, touch_id, surface_x, surface_y
        )

    def touch_send_up(self, time_msec: int, touch_id: int) -> None:
        """
        Send a touch up event for the touch point given by the `touch_id`. The event
        will go to the client for the surface given in the corresponding touch down
        event. This will remove the touch point. This function does not respect touch
        grabs: you probably want `touch_notify_up()` instead.
        """
        lib.wlr_seat_touch_send_up(self._ptr, time_msec, touch_id)

    def touch_send_motion(
        self, time_msec: int, touch_id: int, surface_x: float, surface_y: float
    ) -> None:
        """
        Send a touch motion event for the touch point given by the `touch_id`. The
        event will go to the client for the surface given in the corresponding touch
        down event. This function does not respect touch grabs: you probably want
        `touch_notify_motion()` instead.
        """
        lib.wlr_seat_touch_send_motion(
            self._ptr, time_msec, touch_id, surface_x, surface_y
        )

    def touch_send_cancel(self, surface: Surface) -> None:
        """
        Notify the seat that this is a global gesture and the client should cancel
        processing it. The event will go to the client for the surface given.
        This function does not respect touch grabs: you probably want
        `touch_notify_cancel()` instead.
        """
        lib.wlr_seat_touch_send_cancel(self._ptr, surface._ptr)

    def touch_send_frame(self) -> None:
        lib.wlr_seat_touch_send_frame(self._ptr)

    def touch_notify_down(
        self,
        surface: Surface,
        time_msec: int,
        touch_id: int,
        surface_x: float,
        surface_y: float,
    ) -> int:
        """
        Notify the seat of a touch down on the given surface. Defers to any grab of the
        touch device.

        Returns the serial id.
        """
        return lib.wlr_seat_touch_notify_down(
            self._ptr, surface._ptr, time_msec, touch_id, surface_x, surface_y
        )

    def touch_notify_up(self, time_msec: int, touch_id: int) -> None:
        """
        Notify the seat that the touch point given by `touch_id` is up. Defers to any
        grab of the touch device.
        """
        lib.wlr_seat_touch_notify_up(self._ptr, time_msec, touch_id)

    def touch_notify_motion(
        self, time_msec: int, touch_id: int, surface_x: float, surface_y: float
    ) -> None:
        """
        Notify the seat that the touch point given by `touch_id` has moved. Defers to
        any grab of the touch device. The seat should be notified of touch motion even
        if the surface is not the owner of the touch point for processing by grabs.
        """
        lib.wlr_seat_touch_notify_motion(
            self._ptr, time_msec, touch_id, surface_x, surface_y
        )

    def touch_point_focus(
        self,
        surface: Surface,
        time_msec: int,
        touch_id: int,
        surface_x: float,
        surface_y: float,
    ) -> None:
        """
        Notify the seat that the touch point given by `touch_id` has entered a new
        surface. The surface is required. To clear focus, use touch_point_clear_focus().
        """
        lib.wlr_seat_touch_point_focus(
            self._ptr, surface._ptr, time_msec, touch_id, surface_x, surface_y
        )

    def touch_point_clear_focus(
        self,
        time_msec: int,
        touch_id: int,
    ) -> None:
        """Clear the focused surface for the touch point given by `touch_id`."""
        lib.wlr_seat_touch_point_clear_focus(self._ptr, time_msec, touch_id)

    def touch_notify_cancel(self, surface: Surface):
        """Notify the seat that this is a global gesture and the client should
        cancel processing it. Defers to any grab of the touch device."""
        lib.wlr_seat_touch_notify_cancel(self._ptr, surface._ptr)

    def touch_notify_frame(self) -> None:
        lib.wlr_seat_touch_notify_frame(self._ptr)

    def touch_num_points(self) -> int:
        """
        How many touch points are currently down for the seat.
        """
        return lib.wlr_seat_touch_num_points(self._ptr)

    def touch_has_grab(self) -> bool:
        """
        Whether or not the seat has a touch grab other than the default grab.
        """
        return lib.wlr_seat_touch_has_grab(self._ptr)

    def touch_get_point(self, touch_id: int) -> TouchPoint | None:
        """
        Get the active touch point with the given `touch_id`.

        If the touch point does not exist or is no longer active, returns None.
        """
        ptr = lib.wlr_seat_touch_get_point(self._ptr, touch_id)
        return instance_or_none(TouchPoint, ptr)

    def set_selection(self, source, serial: int) -> None:
        """Sets the current selection for the seat

        None can be provided to clear it.  This removes the previous one if
        there was any. In case the selection doesn't come from a client,
        Display.next_serial() can be used to generate a serial.
        """
        if source is None:
            lib.wlr_seat_set_selection(self._ptr, ffi.NULL, serial)
        else:
            # TODO: wrap source in a data source
            lib.wlr_seat_set_selection(self._ptr, source, serial)

    def set_primary_selection(self, source, serial: int) -> None:
        """Sets the current primary selection for the seat.

        None can be provided to clear it. This removes the previous one if
        there was any. In case the selection doesn't come from a client,
        Display.next_serial() can be used to generate a serial.
        """
        if source is None:
            lib.wlr_seat_set_primary_selection(self._ptr, ffi.NULL, serial)
        else:
            # TODO: wrap source in a data source
            lib.wlr_seat_set_primary_selection(self._ptr, source, serial)

    def validate_pointer_grab_serial(self, origin: Surface, serial: int) -> bool:
        """Check whether this serial is valid to start a pointer grab action."""
        return lib.wlr_seat_validate_pointer_grab_serial(self._ptr, origin._ptr, serial)

    def start_pointer_drag(self, drag: Drag, serial: int) -> None:
        """
        Starts a pointer drag on the seat. This starts implicit keyboard and pointer
        grabs.
        """
        lib.wlr_seat_start_pointer_drag(self._ptr, drag._ptr, serial)

    def surface_accepts_touch(self, surface: Surface) -> bool:
        return lib.wlr_surface_accepts_touch(self._ptr, surface._ptr)

    def __enter__(self) -> Seat:
        """Context manager to clean up the seat"""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        """Clean up the seat when exiting the context"""
        self.destroy()


class PointerRequestSetCursorEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_seat_pointer_request_set_cursor_event *", ptr)

    # TODO: seat client

    @property
    def surface(self) -> Surface:
        # TODO: setup weakref
        return Surface(self._ptr.surface)

    @property
    def serial(self) -> int:
        return self._ptr.serial

    @property
    def hotspot(self) -> tuple[int, int]:
        return self._ptr.hotspot_x, self._ptr.hotspot_y


class RequestSetSelectionEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_seat_request_set_selection_event *", ptr)

    # TODO: source

    @property
    def serial(self) -> int:
        return self._ptr.serial


class RequestSetPrimarySelectionEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast(
            "struct wlr_seat_request_set_primary_selection_event *", ptr
        )

    # TODO: source

    @property
    def serial(self) -> int:
        return self._ptr.serial


class RequestStartDragEvent(Ptr):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_seat_request_start_drag_event *", ptr)

    @property
    def drag(self) -> Drag:
        return Drag(self._ptr.drag)

    @property
    def origin(self) -> Surface:
        return Surface(self._ptr.origin)

    @property
    def serial(self) -> int:
        return self._ptr.serial


class _FocusChangeEvent(Ptr):
    """
    Base class for ...FocusChangeEvents which provides common properties.
    """

    # TODO: wlr_seat *seat

    @property
    def old_surface(self) -> Surface:
        # TODO: May old_surface be NULL?
        return Surface(self._ptr.old_surface)

    @property
    def new_surface(self) -> Surface:
        return Surface(self._ptr.new_surface)


class PointerFocusChangeEvent(_FocusChangeEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_seat_pointer_focus_change_event *", ptr)

    @property
    def surface_x(self) -> float:
        return self._ptr.sx

    @property
    def surface_y(self) -> float:
        return self._ptr.sy


class KeyboardFocusChangeEvent(_FocusChangeEvent):
    def __init__(self, ptr) -> None:
        self._ptr = ffi.cast("struct wlr_seat_keyboard_focus_change_event *", ptr)


class SeatPointerState(Ptr):
    def __init__(self, ptr) -> None:
        """The current state of the pointer on the seat"""
        self._ptr = ptr

        self.focus_change_event = Signal(
            ptr=ffi.addressof(self._ptr.events.focus_change),
            data_wrapper=PointerFocusChangeEvent,
        )

    @property
    def surface_x(self) -> float:
        return self._ptr.sx

    @property
    def surface_y(self) -> float:
        return self._ptr.sy

    @property
    def focused_surface(self) -> Surface | None:
        """The surface that currently has keyboard focus"""
        return instance_or_none(Surface, self._ptr.focused_surface)


class SeatKeyboardState(Ptr):
    def __init__(self, ptr) -> None:
        """The current state of the keyboard on the seat"""
        self._ptr = ptr

        self.focus_change_event = Signal(
            ptr=ffi.addressof(self._ptr.events.focus_change),
            data_wrapper=KeyboardFocusChangeEvent,
        )

    @property
    def focused_surface(self) -> Surface | None:
        """The surface that is currently focused"""
        return instance_or_none(Surface, self._ptr.focused_surface)


class SeatTouchState(Ptr):
    def __init__(self, ptr) -> None:
        """The current state of touch on the seat"""
        self._ptr = ptr

    @property
    def grab_serial(self) -> int:
        return self._ptr.grab_serial

    def grab_id(self) -> int:
        return self._ptr.grab_id

    @property
    def touch_points(self) -> Iterator[TouchPoint]:
        for ptr in wl_list_for_each(
            "struct wlr_touch_point *",
            self._ptr.touch_points,
            "link",
            ffi=ffi,
        ):
            yield TouchPoint(ptr)


class TouchPoint(Ptr):
    def __init__(self, ptr):
        self._ptr = ptr

    @property
    def touch_id(self) -> int:
        return self._ptr.touch_id

    @property
    def surface_x(self) -> float:
        return self._ptr.sx

    @property
    def surface_y(self) -> float:
        return self._ptr.sy

    @property
    def surface(self) -> Surface | None:
        return instance_or_none(Surface, self._ptr.surface)

    @property
    def focused_surface(self) -> Surface | None:
        """The surface that is currently focused

        Note: wlroot calls it "focus_surface" renamed it to "focused_surface"
        to keep the name aligned to "SeatKeyboardState" and "SeatPointerState".
        """
        return instance_or_none(Surface, self._ptr.focus_surface)
